/* This file is part of Cloudy and is copyright (C)1978-2014 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
#include "cddefines.h"
#include "lines.h"
#include "cddrive.h"
#include "prt.h"
#include "lines_service.h"

t_LineSave LineSave;
/* these are the definitions of the line save arrays in lines.h */

void t_LineSave::zero()
{
	DEBUG_ENTRY( "t_LineSave::zero()" );

	/* index within the line in the line stack 
	 * default is Hbeta total - the third line in the stack
	 * 0th is a zero for sanity, 1st is unit, 2nd is a comment */
	/* >>chng 02 apr 22 from 2 to 3 since added unit at 1 */
	/* >>chng 06 mar 11, from 3 to -1 will now set to "H  1" 4861 */
	ipNormWavL = -1;
	WavLNorm = 4861.36f;
	lgNormSet = false;
	sig_figs = 6;

	/* the label for the normalization line */
	strcpy( chNormLab, "    " );

	/* The scale factor for the normalization line */
	ScaleNormLine = 1.;

}

void LinSv::prt(FILE* ioPUN) const
{
	DEBUG_ENTRY( "LinSv::prt()" );
	fprintf(ioPUN,"%s",label().c_str());
}

string LinSv::label() const
{
	DEBUG_ENTRY( "LinSv::label()" );
	string val = chALab();
	val += " ";
	char buf[100];
	sprt_wl(buf, wavelength());
	val += buf;
	return val;
}

string LinSv::biglabel() const
{
	DEBUG_ENTRY( "LinSv::biglabel()" );
	string val = label();
	if (m_tr.associated())
	{
		val += " ";
		val += "[";
		val += (*m_tr.Hi()).chConfig();
		val += "->";
		val += (*m_tr.Lo()).chConfig();
		val += "]";
	}
	return val;
}

bool LinSv::isCat(const char *s) const
{
	DEBUG_ENTRY( "LinSv::isCat()" );
	return strncmp(chALab(),s,strlen(s)) == 0;
}

void LinSv::chALabSet(const char *that)
{
	/* check that null is correct, string overruns have 
	 * been a problem in the past */
	ASSERT( (int) strlen( that )< NCHLAB );
	strncpy(m_chALab,that,NCHLAB-1);
	m_chALab[NCHLAB-1] = '\0';
	strncpy(m_chCLab,that,NCHLAB-1);
	m_chCLab[NCHLAB-1] = '\0';
	caps(m_chCLab);
}

void LinSv::init(char chSumTyp, const char *chComment, const char *label,
					  bool lgAdd, realnum wavelength, const TransitionProxy& tr)
{
	/* first call to stuff lines in array, confirm that label is one of
	 * the four correct ones */
	ASSERT( (chSumTyp == 'c') || (chSumTyp == 'h') || (chSumTyp == 'i') || (chSumTyp == 'r' ) );
	ASSERT(!m_tr.associated());
	/* then save it into array */
	m_chSumTyp = chSumTyp;
	emslinZero();
	m_chComment = chComment;
	chALabSet( label );
	
	if (!lgAdd)
	{
		// number of lines OK, set parameters for first pass
		// negative wavelength means it is just label, possibly not correct
		wavelength = fabs(wavelength);
	}
	SumLineZero();
	m_wavelength = wavelength;
	m_tr = tr;
}

void LinSv::addComponentID(long id)
{
	m_component.push_back(id);
	if (m_component.size() == 1)
		m_chComment += ": ";
	else
		m_chComment += "+";
	if (id > 0)
	{
		m_chComment += "\""+LineSave.lines[id].label()+"\"";
	}
	else
	{
		m_chComment += "N/A";
	}
}
void LinSv::addComponent(const char* species,const double wavelength)
{
	if ( LineSave.ipass == 0 )
	{
		long id = LineSave.findline(species,wlAirVac(wavelength));
		if (id == 0)
			fprintf( ioQQQ, "Error: A component to line blend \"%s\" %.3f was not identified.\n",
						species, wavelength );
		addComponentID(id);
	}
}
void LinSv::makeBlend(const char* chLabel,const double wavelength, const double width)
{
	DEBUG_ENTRY("LinSv::makeBlend()");
	if ( LineSave.ipass == 0 )
	{
		realnum wlo = wlAirVac(wavelength-width),
			whi = wlAirVac(wavelength+width);
			
		/* check that chLabel[4] is null - supposed to be 4 char + end */
		if( chLabel[NCHLAB-1] != '\0' || strlen(chLabel) != NCHLAB-1 )
		{
			fprintf( ioQQQ, " makeBlend called with insane species \"%s\", must be exactly %d characters long.\n",
						chLabel, NCHLAB-1 );
			cdEXIT(EXIT_FAILURE);
		}
		
		char chCARD[INPUT_LINE_LENGTH];
		strcpy( chCARD, chLabel );
		
		/* make sure chLabel is all caps */
		caps(chCARD);/* convert to caps */
		
		/* this replaces tabs with spaces. */
		/* \todo	2 keep this in, do it elsewhere, just warn and bail? */
		for( char *s=chCARD; *s != '\0'; ++s )
		{
			if( *s == '\t' )
			{
				*s = ' ';
			}
		}

		for( long j=1; j < LineSave.nsum; j++ )
		{
			/* use pre-capitalized version of label to be like input chLineLabel */
			const char *chCaps = LineSave.lines[j].chCLab();

			if (LineSave.lines[j].wavelength() >= wlo && 
				 LineSave.lines[j].wavelength() <= whi &&
				 strcmp(chCaps,chCARD) == 0)
			{
				addComponentID(j);
				fprintf( ioQQQ,"Adding \"%s\" to blend\n", 
							LineSave.lines[j].label().c_str() );
			}
		}
	}
}
string LinSv::chComment() const
{
	return m_chComment;
}

static bool wavelength_compare (long a, long b)
{
	LinSv* a1 = &LineSave.lines[a];
	LinSv* b1 = &LineSave.lines[b];
	/* comparison is b-a so we get inverse wavelength order (increasing energy order) */
	if( b1->wavelength() < a1->wavelength() )
		return true;
	else 
		return false;
}

static bool wavelength_compare_realnum (size_t a, realnum wavelength)
{
	/* comparison is b-a so we get inverse wavelength order (increasing energy order) */
	if( wavelength < LineSave.lines[a].wavelength() )
		return true;
	else 
		return false;
}

void t_LineSave::setSortWL()
{
	DEBUG_ENTRY( "t_LineSave::setSortWL()" );
	SortWL.resize((unsigned)nsum);
	for (long nlin=0; nlin < nsum; ++nlin)
	{
		SortWL[nlin] = nlin;
	}
	stable_sort(SortWL.begin(), SortWL.end(), wavelength_compare);
}

long t_LineSave::findline(const char *chLabel, realnum wavelength)
{
	DEBUG_ENTRY( "t_LineSave::findline()" );
	
	ASSERT( nsum >= 0);
	if (nsum == 0)
		return -1;

	bool lgDEBUG = false;
	
	/* check that chLabel[4] is null - supposed to be 4 char + end */
	if( chLabel[NCHLAB-1] != '\0' || strlen(chLabel) != NCHLAB-1 )
	{
		fprintf( ioQQQ, " findline called with insane chLabel (between quotes) \"%s\", must be exactly %d characters long.\n",
					chLabel, NCHLAB-1 );
		return -2;
	}
 
	char chCARD[INPUT_LINE_LENGTH];
	strcpy( chCARD, chLabel );

	/* make sure chLabel is all caps */
	caps(chCARD);/* convert to caps */

	/* this replaces tabs with spaces. */
	/* \todo	2 keep this in, do it elsewhere, just warn and bail? */
	for( char *s=chCARD; *s != '\0'; ++s )
	{
		if( *s == '\t' )
		{
			*s = ' ';
		}
	}

	/* get the error associated with specified significant figures; add
	 * FLT_EPSILON*wavelength to broaden bounds enough to allow for
	 * cancellation error
	 */
	realnum errorwave = WavlenErrorGet( wavelength, LineSave.sig_figs ) + FLT_EPSILON*wavelength, 
		smallest_error=BIGFLOAT,
		smallest_error_w_correct_label=BIGFLOAT;

	if (1)
	{
		// Possibly more 'user friendly' line search mode -- not active
		// NB falls through to previous implementation if ipass < 1, as
		// lines will not yet be sorted
		if (ipass == 1)
		{
			char wlbuf[INPUT_LINE_LENGTH];
			// Need to search for lines within allowed band, and plausible confusions

			// Find position in list of lines
			vector<size_t>::iterator first = 
				lower_bound(SortWL.begin(),SortWL.end(),
								wavelength+errorwave, wavelength_compare_realnum);

			// first is now first line below upper limit
			if (first == SortWL.end())
			{
				sprt_wl(wlbuf,wavelength);
				fprintf(ioQQQ,"Didn't find anything at %s\n",wlbuf);
				cdEXIT(EXIT_FAILURE);
			}

			// look for first line below lower limit -- may be the same as first if there are no matches
			vector<size_t>::iterator second;
			for(second=first; second != SortWL.end(); ++second)
			{
				if (lines[*second].wavelength() < wavelength-errorwave)
					break;
			}
			
			vector<size_t>::iterator found = SortWL.end();
			int nmatch=0;
			realnum dbest = 0.;
			for (vector<size_t>::iterator pos = first; pos != second; ++pos)
			{
				if ( strcmp(lines[*pos].chCLab(),chCARD) == 0 )
				{
					++nmatch;
					realnum dwl = lines[*pos].wavelength()-wavelength;
					if (nmatch >= 2)
					{
						if (nmatch == 2)
						{
							sprt_wl(wlbuf,wavelength);
							fprintf(ioQQQ,"WARNING: multiple matching lines found in search for \"%s\" %s\n",
									  chLabel,wlbuf);
							fprintf(ioQQQ,"WARNING: match 1 is \"%s\" (dwl=%gA)\n", 
									  lines[*found].biglabel().c_str(),lines[*found].wavelength()-wavelength);
						}
						fprintf(ioQQQ,"WARNING: match %d is \"%s\" (dwl=%gA)\n",
								  nmatch, lines[*pos].biglabel().c_str(),dwl);
					}
					if ( found == SortWL.end() )
					{
						found = pos;
						dbest = fabs(lines[*pos].wavelength()-wavelength);
					} 
					else if ( fabs(dwl) < dbest )
					{
						found = pos;
						dbest = fabs(dwl);
					}
				}
			}
			if ( found != SortWL.end())
			{
				if (0)
					fprintf(ioQQQ,"Found %s\n", lines[*found].label().c_str());
				return *found;
			}
			
			sprt_wl(wlbuf,wavelength);
			fprintf(ioQQQ,"WARNING: no exact matching lines found for \"%s\" %s\n",chLabel,wlbuf);
			for (vector<size_t>::iterator pos = first; pos != second; ++pos)
			{
				fprintf(ioQQQ,"WARNING: Line with incorrect label found close \"%s\"\n",
						  lines[*pos].label().c_str());
			}
			// Haven't found a match with correct label
			vector<size_t>::iterator best = SortWL.end();
			realnum besterror = 0.;
			for (;;)
			{
				realnum errordown = lines[*(first-1)].wavelength()-wavelength;
				realnum errorup = wavelength-lines[*second].wavelength();
				realnum error = 0.;
				vector<size_t>::iterator next;
				if ( errordown < errorup )
				{
					error = errordown;
					--first;
					next = first;
				}
				else
				{
					error = errorup;
					next = second;
					++second;
				}
				if ( strcmp(lines[*next].chCLab(),chCARD) == 0 )
				{
					if (best == SortWL.end())
					{
						best = next;
						besterror = error;
						fprintf(ioQQQ,"Taking best match as \"%s\"\n",lines[*next].label().c_str());
					}
					else
					{
						fprintf(ioQQQ,"Possible ambiguity with \"%s\"\n",lines[*next].label().c_str());
					}
				}
				// Assume this is clearly unambiguous
				if (best != SortWL.end() && error > 100.*besterror)
					break;
				// Assume this is clearly unmatched
				if (error > 0.01*wavelength)
					break;
			}
			if (best != SortWL.end())
				return *best;
			fprintf(ioQQQ,"PROBLEM: no matching line found\n");
			cdEXIT(EXIT_FAILURE);
		}
		// Falls through to previous implementation if ipass < 1.  Can
		// be more restrictive with match handling, as this will be from
		// a request internal to the code infrastructure, rather than
		// user data
	}

	long int j, index_of_closest=LONG_MIN,
		index_of_closest_w_correct_label=-1;
	
	for( j=1; j < nsum; j++ )
	{
		realnum current_error = (realnum)fabs(lines[j].wavelength()-wavelength);
		/* use pre-capitalized version of label to be like input chLineLabel */
		const char *chCaps = lines[j].chCLab();

		if( current_error < smallest_error )
		{
			index_of_closest = j;
			smallest_error = current_error;
		}
			
		if( current_error < smallest_error_w_correct_label &&
			 (strcmp(chCaps,chCARD) == 0) )
		{
			index_of_closest_w_correct_label = j;
			smallest_error_w_correct_label = current_error;
		}

		/* check wavelength and chLabel for a match */
		/*if( fabs(lines[j].wavelength- wavelength)/MAX2(DELTA,wavelength)<errorwave && 
			strcmp(chCaps,chCARD) == 0 ) */
		if( lgDEBUG && (current_error <= errorwave || 
			  fp_equal( wavelength + errorwave, lines[j].wavelength() ) ||
			  fp_equal( wavelength - errorwave, lines[j].wavelength() ))  
			 && strcmp(chCaps,chCARD) == 0 )
		{
			/* match, so set emiss to emissivity in line */
			/* and announce success by returning line index within stack */
			printf("Matched %s %15.8g %ld %18.11g %s\n",
					 chCaps,wavelength,j,lines[j].wavelength(),
					 lines[j].biglabel().c_str());
		}
	}




	// Didn't find line, handle error
	if( index_of_closest_w_correct_label == -1 ||
		smallest_error_w_correct_label > errorwave )
	{
		/* >>chng 05 dec 21, report closest line if we did not find exact match, note that
		 * exact match returns above, where we will return negative number of lines in stack */
		fprintf( ioQQQ," PROBLEM findline did not find line " );
		prt_line_err( ioQQQ, chCARD, wavelength );
		if( index_of_closest >= 0 )
		{
			fprintf( ioQQQ,"  The closest line (any label) was   \"%s\"\n", 
						lines[index_of_closest].label().c_str() );
			if( index_of_closest_w_correct_label >= 0 )
			{
				fprintf( ioQQQ,"  The closest with correct label was \"%s\"\n", 
							lines[index_of_closest_w_correct_label].label().c_str() );
				fprintf( ioQQQ,"  Error was %15.8g vs. tolerance %15.8g\n", 
							smallest_error_w_correct_label, errorwave );
			}
			else
				fprintf( ioQQQ,"\n  No line found with label \"%s\".\n", chCARD );
			fprintf( ioQQQ,"\n" );
		}
		else
		{
			fprintf( ioQQQ,".\n PROBLEM No close line was found\n" );
			TotalInsanity();
		}
		return -3;
	}

	if (lgDEBUG)
		fprintf(ioQQQ,"Identified %ld\n",index_of_closest_w_correct_label);


	return index_of_closest_w_correct_label;
}

/*************************************************************************
 *
 * cdEmis obtain the local emissivity for a line with known index
 *
 ************************************************************************/
void cdEmis(
	const LinSv* line,
	/* the vol emissivity of this line in last computed zone */
	double *emiss ,
	// intrinsic or emergent
	bool lgEmergent )
{
	DEBUG_ENTRY( "cdEmis()" );

	if (line)
		*emiss = line->emslin(lgEmergent);
	else
		*emiss = 0.;
}

