/* This file is part of Cloudy and is copyright (C)1978-2014 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
/*SaveSpecies generate output for the save species command */
#include "cddefines.h"
#include "taulines.h"
#include "radius.h"
#include "save.h"
#include "mole.h"
#include "mole_priv.h"

/* save results for one particular species */
STATIC void SaveSpeciesOne( const molecule* spg, const char chKey[],
		FILE *ioPUN, long int ipPun, size_t maxLevels );

/*SaveSpecies generate output for the save species command */
void SaveSpecies(
		FILE* ioPUN,
		long int ipPun)
{
	DEBUG_ENTRY( "SaveSpecies()" );

	/* first branch; save results for all species if list is empty */
	vector<const molecule *> speciesList;
	if( save.chSaveSpecies[ipPun].size() == 0 )
	{
		// loop over species
		for( size_t i=0; i<mole_global.list.size(); ++i )
		{
			speciesList.push_back( &(*mole_global.list[i]) );
		}
	}
	else
	{
		for (vector<string>::iterator name=save.chSaveSpecies[ipPun].begin();
			  name != save.chSaveSpecies[ipPun].end(); ++name)
		{
			const molecule *saveSpeciesGlobal = findspecies(name->c_str());
			if( saveSpeciesGlobal == null_mole )
			{
				fprintf( ioQQQ,"Could not find species '%s', "
							"use SAVE SPECIES LABELS to get a list of all species."
							"\nSorry.\n", name->c_str() );
				cdEXIT(EXIT_FAILURE);
			}
			speciesList.push_back(saveSpeciesGlobal);
		}
	}

	if( strcmp( save.chSaveArgs[ipPun], "LABE" )==0 )
	{
		if( save.lgSaveHeader(ipPun) )
		{
			/* save list of species labels */
			fprintf( ioPUN, "#Species labels\n" );
			for( size_t i=0; i<speciesList.size(); ++i )
			{
				fprintf( ioPUN, "%s\n", speciesList[i]->label.c_str() );
			}
			save.SaveHeaderDone(ipPun);
		}
		return;
	}

	else if( strcmp(save.chSaveArgs[ipPun],"LEVL") == 0 )
	{
		/* number of levels active in this zone */
		if( save.lgSaveHeader(ipPun) )
		{
			/* save list of species labels */
			fprintf( ioPUN, "#Species\tnumber of levels\n" );
			save.SaveHeaderDone(ipPun);
		}
		for( size_t i=0; i<speciesList.size(); ++i )
		{
			const molecule *spg = speciesList[i];
			const molezone *sp = spg->local();
			fprintf( ioPUN, "%s", spg->label.c_str() );
			if( sp->levels == NULL )
				fprintf( ioPUN, "\t%4lu\n", 0UL );
			else
				fprintf( ioPUN, "\t%4lu\n", (unsigned long)sp->levels->size() );
		}
		return;
	}
	
	/* remaining options are column densities, populations, and energies */
	// max number of levels, for header print
	size_t mostLevels = 0;
	for (vector<const molecule *>::iterator sp=speciesList.begin();
		  sp != speciesList.end(); ++sp)
	{
		const molezone *saveSpecies = (*sp)->local();
		if( saveSpecies->levels != NULL )		
			mostLevels = MAX2(mostLevels, saveSpecies->levels->size() );
	}
	
	ASSERT( mostLevels < 10000 );
	for (vector<const molecule *>::iterator sp=speciesList.begin();
		  sp != speciesList.end(); ++sp)
	{
		SaveSpeciesOne( *sp, save.chSaveArgs[ipPun], ioPUN, ipPun, mostLevels );
	}

	return;
}

/* print 0.000e+00 as simply 0 */
STATIC void PrintShortZero( FILE *ioPUN , double arg )
{
	DEBUG_ENTRY( "PrintShortZero()" );
	if( arg==0. )
		fprintf(ioPUN,"\t0");
	else
		fprintf(ioPUN,"\t%.3e", arg);

}

/* save results for one particular species */
STATIC void SaveSpeciesOne( const molecule* spg, const char chKey[],
		FILE *ioPUN, long int ipPun, size_t maxLevels )
{
	DEBUG_ENTRY( "SaveSpeciesOne()" );

	molezone *sp = spg->local();

	if( spg == null_mole || sp == null_molezone )
		return;
	
	// one time print of energy levels
	if( strcmp( chKey, "ENER" )==0 )
	{
		if( save.lgSaveHeader(ipPun) )
		{
			fprintf( ioPUN, "#species energies");
			for( size_t i = 0; i < maxLevels; ++i )
			{
				fprintf( ioPUN, "\t%lu", (unsigned long)i );
			}
			fprintf( ioPUN, "\n");
			save.SaveHeaderDone(ipPun);
		}

		fprintf( ioPUN, "%s", spg->label.c_str() );
		if( sp->levels == NULL || sp->levels->size() == 0 )
		{
			fprintf( ioPUN, "\t%.6e", 0. );
		}
		else
		{
			for( qList::const_iterator st = sp->levels->begin(); st != sp->levels->end(); ++st )
			{
				ASSERT( (*st).g() > 0.f );
				fprintf( ioPUN, "\t%.6e", AnuUnit( (*st).energy().Ryd() ) );
			}
		}
		fprintf( ioPUN, "\n");
		return;
	}

	if( strcmp( chKey, "DEPA" )==0 )
	{
		if( save.lgSaveHeader(ipPun) )
		{
			fprintf( ioPUN, "#depth [cm] species departure coefficients [dimensionless]");

			for( size_t i = 0; i < maxLevels; ++i )
			{
				fprintf( ioPUN, "\t%lu", (unsigned long)i );
			}
			fprintf( ioPUN, "\n");
			save.SaveHeaderDone(ipPun);
		}

		fprintf( ioPUN, "%.5e", radius.depth_mid_zone );
		fprintf( ioPUN, "\t%s", spg->label.c_str() );

		if( sp->levels == NULL || sp->levels->size() == 0 )
		{
			// just print unity if no internal structure available
			PrintShortZero( ioPUN, 1.0 );
		}
		else
		{
			bool lgZeroHit = false;
			// loop over levels for this species
			for( qList::const_iterator st = sp->levels->begin(); st != sp->levels->end(); ++st )
			{
				// don't print high levels which can have 0 abundance at low temperature
				fixit("probably want to rethink whether we want to avoid printing zeroes here.");
				if( !lgZeroHit )
					PrintShortZero( ioPUN, (*st).DepartCoef() );
				if( (*st).DepartCoef() == 0.)
					lgZeroHit = true;
			}
		}
		fprintf( ioPUN, "\n");
	}
	else if( strcmp( chKey, "POPU" )==0 )
	{
		if( save.lgSaveHeader(ipPun) )
		{
			fprintf( ioPUN, "#depth [cm] species populations [cm-3]");

			for( size_t i = 0; i < maxLevels; ++i )
			{
				fprintf( ioPUN, "\t%lu", (unsigned long)i );
			}
			fprintf( ioPUN, "\n");
			save.SaveHeaderDone(ipPun);
		}

		fprintf( ioPUN, "%.5e", radius.depth_mid_zone );
		fprintf( ioPUN, "\t%s", spg->label.c_str() );

		if( sp->levels == NULL || sp->levels->size() == 0 )
		{
			PrintShortZero( ioPUN, sp->den );
		}
		else
		{
			bool lgZeroHit = false;
			// loop over levels for this species
			for( qList::const_iterator st = sp->levels->begin(); st != sp->levels->end(); ++st )
			{
				// don't print high levels which can have 0 abundance at low temperature
				if( !lgZeroHit )
					PrintShortZero( ioPUN, (*st).Pop() );
				if( (*st).Pop() == 0.)
					lgZeroHit = true;
			}
		}
		fprintf( ioPUN, "\n");
	}
	else if( strcmp( chKey, "COLU" )==0 )
	{
		if( save.lgSaveHeader(ipPun) )
		{
			fprintf( ioPUN, "#species column density [cm-2]");

			for( size_t i = 0; i < maxLevels; ++i )
			{
				fprintf( ioPUN, "\t%lu", (unsigned long)i );
			}
			fprintf( ioPUN, "\n");
			save.SaveHeaderDone(ipPun);
		}

		fprintf( ioPUN, "%s", spg->label.c_str() );

		if( sp->levels == NULL || sp->levels->size() == 0 )
		{
			PrintShortZero( ioPUN, sp->column );
		}
		else
		{
			// loop over levels
			bool lgZeroHit = false;
			for( qList::const_iterator st = sp->levels->begin(); st != sp->levels->end(); ++st )
			{
				// don't print high levels which can have 0 abundance at low temperature
				if( !lgZeroHit )
					PrintShortZero( ioPUN, (*st).ColDen() );
				if( (*st).ColDen() == 0.)
					lgZeroHit = true;
			}
		}
		fprintf( ioPUN, "\n");
	}
	
	return;
}

class Field
{
public:
	const char* label;
	const char* fmt;
private:
	double m_value;
public:
	Field(const char* label, const char* fmt, double value) :
		label(label), fmt(fmt), m_value(value)
		{}
	double value() const
		{
			return m_value;
		}
};

STATIC void doHeader(FILE *punit, const Field& f)
{
	DEBUG_ENTRY( "doHeader()" );
	fprintf(punit,"%s",f.label);
}
STATIC void doData(FILE *punit, const Field& f)
{
	DEBUG_ENTRY( "doData()" );
	fprintf(punit,f.fmt,f.value());
}

STATIC inline bool isCatalystReactant(const mole_reaction& rate, int i)
{
	return rate.rvector[i]!=NULL;
}
STATIC inline bool isCatalystProduct(const mole_reaction& rate, int i)
{
	return rate.pvector[i]!=NULL;
}
STATIC inline bool isDestroyed(const mole_reaction& rate, int i)
{
	return !isCatalystReactant(rate,i) && rate.rvector_excit[i]==NULL;
}
STATIC inline bool isCreated(const mole_reaction& rate, int i)
{
	return !isCatalystProduct(rate,i) && rate.pvector_excit[i]==NULL;
}

/* Save all rates involving specified species */
void mole_save(FILE *punit, const char speciesname[], const char args[], bool lgHeader, bool lgData, bool lgCoef, double depth)
{
	DEBUG_ENTRY( "mole_save()" );

	molecule *sp = findspecies(speciesname);
	
	if( sp == null_mole )
	{
		fprintf( punit, " Species %s could not be found. Note that labels are case-sensitive in this context.\n", speciesname );
		return;
	}
		
	void (*doTask)(FILE *punit, const Field& f);

	if (lgHeader)
		doTask=doHeader;
	else
		doTask=doData;
	
	doTask(punit,Field("Depth","%.5e",depth));

	for (mole_reaction_i p
				 =mole_priv::reactab.begin(); p != mole_priv::reactab.end(); ++p) 
	{
		const mole_reaction &rate = *p->second;
		bool ipthis = false;

		for (int i=0;i<rate.nreactants && !ipthis;i++)
		{
			if( rate.reactants[i] == sp )
			{
				if( ( strcmp( args, "DEST" )==0 && isDestroyed(rate,i) ) ||
					 ( strcmp( args, "DFLT" )==0 && isDestroyed(rate,i) ) ||
					 ( strcmp( args, "CATA" )==0 && isCatalystReactant(rate,i) ) ||
					strcmp( args, "ALL " )==0 )
					ipthis = true;
			}
		}
		
		for(int i=0;i<rate.nproducts && !ipthis;i++)
		{
			if( rate.products[i] == sp )
			{
				if( ( strcmp( args, "CREA" )==0 && isCreated(rate,i) ) ||
					( strcmp( args, "DFLT" )==0 && isCreated(rate,i) ) ||
					( strcmp( args, "CATA" )==0 && isCatalystProduct(rate,i) ) ||
					strcmp( args, "ALL " )==0 )
					ipthis = true;
			}
		}

		if(ipthis) 
		{
			double ratevi=0.0;
			if (lgData)
			{
				ratevi = mole.reaction_rks[ rate.index ];
				if( !lgCoef )
				{
					for(int i=0;i<rate.nreactants;i++)
					{
						ratevi *= mole.species[ rate.reactants[i]->index ].den;
					}
				}
			}
			fprintf(punit,"\t");

			doTask(punit,Field(rate.label.c_str(),"%.3e",ratevi));
		}
	}
	fprintf(punit,"\n");
}	

// Helper class for sorting rates
class RateCmp
{
	const vector<double>& m_stack;
public:
	RateCmp(const vector<double>& stack) : m_stack(stack) {}
	bool operator()(size_t a, size_t b)
		{
			return m_stack[a] > m_stack[b];
		}
};

void mole_dominant_rates( const vector<const molecule*>& debug_list, 
								  FILE *ioOut,
								  bool lgPrintReagents, size_t NPRINT, double fprint )
{
	vector<double> snkx, srcx;
	vector<mole_reaction *> ratesnk, ratesrc;

	double src_total = 0.0, snk_total = 0.0;

	for(mole_reaction_i p=mole_priv::reactab.begin();
			p != mole_priv::reactab.end(); ++p)
	{
		mole_reaction *rate = &(*p->second);
		double rk = mole.reaction_rks[ rate->index ];

		double rate_tot = rk;
		for(long i=0;i<rate->nreactants;i++)
		{
			rate_tot *= mole.species[ rate->reactants[i]->index ].den;
		}
		
		int nrate=0;
		for(size_t s=0; s<debug_list.size(); ++s)
		{
			for(long i=0;i<rate->nproducts;++i)
			{
				molecule *sp = rate->products[i];
				if (sp == debug_list[s] && rate->pvector[i] == NULL)
				{
					++nrate;
				}
			}
			for(long i=0;i<rate->nreactants;++i)
			{
				molecule *sp = rate->reactants[i];
				if (sp == debug_list[s] && rate->rvector[i] == NULL)
				{
					--nrate;
				}
			}
		}
		if (nrate > 0)
		{
			srcx.push_back(nrate*rate_tot);
			ratesrc.push_back(rate);
			src_total += nrate*rate_tot;
		}
		else if (nrate < 0)
		{
			snkx.push_back(-nrate*rate_tot);
			ratesnk.push_back(rate);
			snk_total -= nrate*rate_tot;
		}
	}

	if (!ratesrc.empty())
	{
		vector<size_t> isrc(ratesrc.size());
		for (size_t i=0; i<ratesrc.size(); ++i)
			isrc[i] = i;
		sort(isrc.begin(),isrc.end(),RateCmp(srcx));

		fprintf( ioOut, "Src %13.7g: ",src_total);
		for (size_t i=0; i<ratesrc.size(); ++i)
		{
			if (i == NPRINT || srcx[isrc[i]] < fprint*src_total)
			{
				break;
			}
			fprintf( ioOut, "%20.20s %13.7g",
						ratesrc[isrc[i]]->label.c_str(),srcx[isrc[i]]);
			if (lgPrintReagents)
			{
				fprintf( ioOut, " [");
				for (long j=0;j<ratesrc[isrc[i]]->nreactants;j++)
				{
					if (j)
					{
						fprintf( ioOut, "," );
					}
					fprintf( ioOut, "%-6.6s %13.7g",
								ratesrc[isrc[i]]->reactants[j]->label.c_str(),
								mole.species[ ratesrc[isrc[i]]->reactants[j]->index ].den);
				}
				fprintf( ioOut, "]" );
			}
			else
			{
				fprintf( ioOut, ";" );
			}
		}
	}
	if (!ratesnk.empty())
	{		
		vector<size_t> isnk(ratesnk.size());
		for (size_t i=0; i<ratesnk.size(); ++i)
			isnk[i] = i;
		sort(isnk.begin(),isnk.end(),RateCmp(snkx));

		fprintf( ioOut, " Snk %13.7g: ", snk_total);
		for (size_t i=0; i<ratesnk.size(); ++i)
		{
			if (i == NPRINT || snkx[isnk[i]] < fprint*snk_total)
			{
				break;
			}
			fprintf( ioOut, "%20.20s %13.7g",
						ratesnk[isnk[i]]->label.c_str(), snkx[isnk[i]] );
			if (lgPrintReagents)
			{
				fprintf( ioOut, " [" );
				for (long j=0;j<ratesnk[isnk[i]]->nreactants;j++)
				{
					if (j)
					{
						fprintf( ioOut, "," );
					}
					fprintf( ioOut, "%-6.6s %13.7g",
								ratesnk[isnk[i]]->reactants[j]->label.c_str(),
								mole.species[ ratesnk[isnk[i]]->reactants[j]->index ].den);
				}
				fprintf( ioOut, "]" );
			}
			else
			{
				fprintf( ioOut, ";" );
			}
		}
	}
	fprintf( ioOut, "\n" );

	return;
}

void mole_print_species_reactions( molecule *speciesToPrint )
{
	if( speciesToPrint==NULL )
	{	
		fprintf( ioQQQ,"\n NULL species found in mole_print_species_reactions.\n" );
		return;
	}

	long numReacts = 0;

	fprintf( ioQQQ,"\n Reactions involving species %s:\n", speciesToPrint->label.c_str() );
	for( mole_reaction_i p=mole_priv::reactab.begin(); p!=mole_priv::reactab.end(); ++p )
	{
		mole_reaction &rate = *p->second;
		for( long i=0; i<rate.nreactants; i++ )
		{
			molecule *sp = rate.reactants[i];
			if(rate.rvector[i]==NULL && sp==speciesToPrint )
			{
				double drate = mole.reaction_rks[ rate.index ];
				for (long j=0; j<rate.nreactants; ++j)
				{
					drate *= mole.species[ rate.reactants[j]->index ].den;
				}
				fprintf( ioQQQ,"%s rate = %g\n", rate.label.c_str(), drate );	
				numReacts++;
			}
		}
		for( long i=0; i<rate.nproducts; i++ )
		{
			molecule *sp = rate.products[i];
			if(rate.pvector[i]==NULL && sp==speciesToPrint )
			{
				double drate = mole.reaction_rks[ rate.index ];
				for (long j=0; j<rate.nreactants; ++j)
				{
					drate *= mole.species[ rate.reactants[j]->index ].den;
				}
				fprintf( ioQQQ,"%s rate = %g\n", rate.label.c_str(), drate );	
				numReacts++;
			}
		}
	}
	fprintf( ioQQQ," End of reactions involving species %s.  There were %li.\n", speciesToPrint->label.c_str(), numReacts );
	
	return;
}
